(**************************************************************************

   	This file is part of the AtomicDensityMatrix package for Mathematica
    Copyright (C) 2004-2019 Simon M. Rochester <simon@rochesterscientific.com>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

**************************************************************************)

BeginPackage["AtomicDensityMatrix`Common`", {"AtomicDensityMatrix`"}];

If[$VersionNumber < 7.0, 
	PauliMatrix::usage = "";
	GatherBy::usage = "";
	Gather::usage = "";
];

If[$VersionNumber < 8.0,
    Block[{$ContextPath}, Quiet@Needs["Combinatorica`"]];
    InversePermutation = Combinatorica`InversePermutation
];

If[$VersionNumber < 9.0,
    ArrayReshape::usage = ""
];

If[$VersionNumber < 10.0,
    SubsetQ::usage = ""
];

If[$VersionNumber < 10.1,
    Subdivide::usage = ""
];

If[$VersionNumber < 10.2,
    ListStepPlot::usage = "";
    EulerMatrix::usage = ""
];

SampleExpression::usage = "";

$Sample::usage = "";

processGraphics::usage = "";

Graphics$::usage = "";

Show$::usage = "";

UnitTestInit::usage = "";

Begin["`Private`"];

SetAttributes[QuietCheck, HoldAll];
QuietCheck[expr_, failexpr_, msgs_] := Quiet[Check[expr, failexpr, msgs], msgs];

FalseQ[expr_] := TrueQ[!expr];

HalfIntegerQ[x_] := IntegerQ[2x];

FunctionDefinedQ[f_, x_] := ValueQ[f[x]];

BlockDiagonal[m : {__?MatrixQ}] :=
    Module[{d = Dimensions /@ m},
        ArrayFlatten@Table[
            If[ i === j, m[[i]], Table[0, {d[[i, 1]]}, {d[[j, 2]]}]],
            {i, Length[m]}, {j, Length[m]}
        ]
    ];

Slope[{{x1_, y1_}, {x2_, y2_}}] := If[x2 != x1, (y2 - y1) / (x2 - x1), Infinity];

If[
    $VersionNumber < 7.0,
    PauliMatrix[0] = {{1, 0}, {0, 1}};
    PauliMatrix[1] = {{0, 1}, {1, 0}};
    PauliMatrix[2] = {{0, -I}, {I, 0}};
    PauliMatrix[3] = {{1, 0}, {0, -1}};
    PauliMatrix[4] = {{1, 0}, {0, 1}};
    SetAttributes[PauliMatrix, Listable];
];

If[$VersionNumber < 7.0,
	GatherBy[list_List, {f1_, f2_}] := Map[GatherBy[#, f2] &, GatherBy[list, f1]];
	GatherBy[list_List, {f_}] := GatherBy[list, f];
	GatherBy[list_List, f_] := Gather[list, f[#1] === f[#2] &];
	Gather[list_List, test_] := Module[{tally},
   		tally = Tally[list, test][[All, 1]];
   		Table[Select[list, test[item, #] &], {item, tally}]
   	];
   	Gather[list_List] := Gather[list, SameQ]
];

If[$VersionNumber < 9.0,
	ArrayReshape[list_, dims_] := Fold[Partition, Flatten[list], dims[[-1 ;; 2 ;; -1]]]
];

If[$VersionNumber < 10.0,
	SubsetQ[set_, subset_] := And @@ (MemberQ[set, #] & /@ subset);
];

If[$VersionNumber < 10.1,
	Subdivide[n_Integer] := Range[0, n]/n;
	Subdivide[xmax_, n_Integer] := xmax Subdivide[n];
	Subdivide[xmin_, xmax_, n_Integer] := Expand@Subdivide[xmax - xmin, n] + xmin;
];

If[$VersionNumber < 10.2,
	ListStepPlot[data_, opts : OptionsPattern[]] := ListLinePlot[data, opts, InterpolationOrder -> 0];
	ListStepPlot[data_, Left | Right | Center, opts : OptionsPattern[]] := ListLinePlot[data, opts, InterpolationOrder -> 0];
];

If[$VersionNumber < 10.2,
    EulerMatrix[{alpha_, beta_, gamma_}, {i_, j_, k_}] :=
        RotationMatrix[alpha, UnitVector[3, i]].RotationMatrix[beta, UnitVector[3, j]].RotationMatrix[gamma, UnitVector[3, k]];
    EulerMatrix[{alpha_, beta_, gamma_}] := EulerMatrix[{alpha, beta, gamma}, {3, 2, 3}];
];

If[$VersionNumber < 12.2,
    Unprotect[Transpose];
    Transpose[{}] = {};
    Protect[Transpose];
];

(* https://mathematica.stackexchange.com/a/154287 *)
If[$VersionNumber < 13.0,
    SparseReplaceAll[s_SparseArray, rule_] := With[
        {elem = Replace[s["NonzeroValues"], rule, Infinity], def = s["Background"] /. rule},
        Replace[s, Verbatim[SparseArray][a__, _, {b__, _}] :> SparseArray[a, def, {b, elem}]]
    ],
    SparseReplaceAll[s_SparseArray, rule_] := SparseArray[
        Automatic,
        Dimensions[s],
        s["Background"] /. rule,
        {
            1,
            {s["RowPointers"], s["ColumnIndices"]},
            Replace[s["NonzeroValues"], rule, Infinity]
        }
    ]
];

SparseReplaceAll[expr_, rule_] := expr /. rule /. s_SparseArray :> SparseReplaceAll[s, rule];

EnsureList[expr_] := If[ListQ@expr, expr, {expr}];

AllSameQ[list_, expr_] := And @@ (expr === # & /@ list);

(*ReplaceAt[expr_,*)
(*    rules : ((_Rule | _RuleDelayed) | {(_Rule | _RuleDelayed) ...}),*)
(*    pos : {},*)
(*    levelspec : (_Integer | Infinity | {_Integer | Infinity} | {(_Integer | Infinity), (_Integer | Infinity)}) : {0},*)
(*    opts : OptionsPattern[]*)
(*] := Replace[expr, Flatten[{rules}], levelspec, opts];*)

(*ReplaceAt[expr_,*)
(*    rules : ((_Rule | _RuleDelayed) | {(_Rule | _RuleDelayed) ...}),*)
(*    pos : ({__Integer} | _Integer),*)
(*    levelspec : (_Integer | Infinity | {_Integer | Infinity} | {(_Integer | Infinity), (_Integer | Infinity)}) : {0},*)
(*    opts : OptionsPattern[]*)
(*] := MapAt[Replace[#, Flatten[{rules}], levelspec, opts] &, expr, Flatten[{pos}]];*)

(*ReplaceAt[expr_,*)
(*    rules : ((_Rule | _RuleDelayed) | {(_Rule | _RuleDelayed) ...}),*)
(*    pos : {__List},*)
(*    levelspec : (_Integer | Infinity | {_Integer | Infinity} | {(_Integer | Infinity), (_Integer | Infinity)}) : {0},*)
(*    opts : OptionsPattern[]*)
(*] := Fold[ReplaceAt[#1, rules, #2, levelspec, opts] &, expr, pos];*)

ReplaceAllAt[expr_,
    rules : ((_Rule | _RuleDelayed) | {(_Rule | _RuleDelayed) ...}),
    pos : {}
] := ReplaceAll[expr, Flatten[{rules}]];

ReplaceAllAt[expr_,
    rules : ((_Rule | _RuleDelayed) | {(_Rule | _RuleDelayed) ...}),
    pos : ({__Integer} | _Integer)
] := MapAt[ReplaceAll[#, Flatten[{rules}]] &, expr, Flatten[{pos}]];

ReplaceAllAt[expr_,
    rules : ((_Rule | _RuleDelayed) | {(_Rule | _RuleDelayed) ...}),
    pos : {__List}
] := Fold[ReplaceAllAt[#1, rules, #2] &, expr, pos];

(* Used in MUnit tests to avoid tests failing due to differing round-off error on different machines *)
RationalizeMachineNumber[x_Real?MachineNumberQ, prec_] :=
    #1 10^#2 & @@ Round[MantissaExponent[x], 10^-prec];

RationalizeMachineNumber[x_Integer, prec_] := x;

RationalizeMachineNumbers[expr_, prec_ : 3] :=
    expr /. {
        x_Real?MachineNumberQ :> RationalizeMachineNumber[x, prec],
        x_Complex /; (MachineNumberQ[x] || MachineNumberQ[x / I]) :> RationalizeMachineNumber[Re@x, prec] + I RationalizeMachineNumber[Im@x, prec]
    };

processGraphics[expr_] :=
    Replace[expr /.
        {InterpolatingFunction[__] -> InterpolatingFunction, HoldPattern[DynamicModule[__]] -> Manipulate, Short[__] -> Short, Manipulate[__] -> Manipulate, Graphics3D[__] -> Graphics3D, _Image | _Graphics | _Graphics$ -> Graphics},
        x_ /; ByteCount@InputForm@x > 1048576/4 -> "BigCell"
    ];

SampleExpression[expression_] := Module[{bytes, expr, pruned},
    expr = RationalizeMachineNumbers[expression /. a_SparseArray :> ArrayRules[a]];
    If[ByteCount@expr < 10000,
        expr,
        If[Length@expr > 1,
            expr = Sort[List @@ expr, ByteCount@#1 > ByteCount@#2 &]
        ];
        bytes = ByteCount@First@expr;
        pruned = Join[
            {First@expr},
            TakeWhile[Rest@expr, (bytes += ByteCount@#) < 10000&]
        ];
        $Sample[
            SampleExpression /@ pruned,
            Dimensions@expr,
            Hash@expr
        ]
    ]
];

UnitTestInit[] := (
    SetOptions[DensityMatrix,TimeDependence->True,Representation->"Zeeman",DMSymbol->\[Rho],DMLabel->None,ComplexExpandVariables->False,TimeVariable->t,AllowedElements->All,NeglectedElements->None];
    SetOptions[PlotAndFit, "ForceNoninteractive" -> True]
);

genericOptionsPattern[1] = _?(MatchQ[Flatten[{#}], {(_Rule | _RuleDelayed) ...}] &);

genericOptionsPattern[] = ___?(MatchQ[Flatten[{#}], {(_Rule | _RuleDelayed) ...}] &);

ZeroMatrix[n_Integer] := ConstantArray[0, {n, n}];
ZeroMatrix[n_Integer, SparseArray] := SparseArray[{}, {n, n}];

EvaluatedFunction[args_, body_] := Function[args, body];

Options[PerturbativeLinearSolve] = {StepFunction -> Simplify};

PerturbativeLinearSolve[eqs : {_Equal ..}, vars_, {lambda_, n_}, OptionsPattern[]] :=
    Module[{A0, b0, A1, b1, stepFunction = OptionValue@StepFunction},
        {{b0, A0}, {b1, A1}} = CoefficientArrays[#, vars] & /@ Flatten /@ Normal@CoefficientArrays[eqs, lambda];
        Thread[vars -> Nest[stepFunction@LinearSolve[A0, -b0 - lambda A1 . #] &, stepFunction@LinearSolve[A0, -b0], n]]
    ];

(* SmartThread: https://mathematica.stackexchange.com/a/23396 *)

SetAttributes[SmartThread, HoldAll];

SmartThread[f_[a_?ArrayQ, b_?ArrayQ], dir : -1 | 1 : -1] := Module[
    {da, db, or, o, g, t},
    or = Function[{x, y, d}, Select[Permutations@Range@Length[x], x[[#[[;; Length[y]]]]] == y &][[d]] ~Check~ Return[$Failed, Module]];
    t = #1 ~Transpose~ Ordering[#2] &;
    g = If[MemberQ[Attributes[f], Listable], f, Function[Null, f[##], Listable]];
    {da, db} = Dimensions /@ {a, b};
    If[Length[da] >= Length[db],
        t[a, o = or[da, db, dir]] ~g~ b,
        a ~g~ t[b, o = or[db, da, dir]]] ~Transpose~ o
];

End[];

EndPackage[];
