BeginPackage["GrotrianDiagram`"];


RelativeCallout[dat_, expr_, pos_] := Callout[dat, expr, dat + pos]
RelativeCallout[dat_, expr_, pos_, apos_] := Callout[dat, expr, dat + pos, dat + apos]
selectEnergyLevels[levelset_, energies_] := Join @@ Table[levelset[Select[#Energy == x&]], {x, energies}]
transitionArrows[levels_, offsets_ : 0] := Module[
    {pos = Normal@levels[All, levelConfigPos]},
    Arrow /@ Partition[pos + offsets, 2, 1]
]
transitionArrows[levels_, offsets_, textpos_] := Module[
    {pos = Normal@levels[All, levelConfigPos]},
    {
        Arrow /@ Partition[pos + offsets, 2, 1],
        MapThread[Style[Text[Row@{#1, " nm"}, #2, #3], FontFamily -> "STIXGeneral", 9]&, {10^4 / Abs@Differences[pos[[All, 2]]], MovingAverage[pos, 2], textpos}]
    }
]
transitionArrow[levels_, e1_ -> e2_, startal_ : 0, endal_ : 0, textal_ : 0] := Module[{level1, level2, pos1, pos2, len1, len2},
    {level1, level2} = Table[levels[Select[#Energy == e&]], {e, {e1, e2}}];
    {pos1, pos2} = Normal /@ Through[{level1, level2}[1, levelConfigPos]];
    {len1, len2} = Normal /@ Through[{level1, level2}[1, {levelLength@#, 0}&]];
    {pos1, pos2} = {pos1 + startal len1, pos2 + endal len2} /. {Left -> 0, Right -> 1, Center -> .5};
    {If[Last@pos2 < Last@pos1, Directive[Gray, Dashed], Style[Text[Row@{NumberForm[10^4 / (Last@pos2 - Last@pos1), {4, 1}], " nm"}, Mean[{pos1, pos2}], {textal, 0} /. {Left -> 1.2, Right -> -1.2, Center -> 0}], FontFamily -> "STIXGeneral", 9]],
        Arrow[{pos1, pos2}]
    }
]
configurationCallouts[levels_, pos_] := Module[{dat},
    MapThread[
        Callout[
            dat = (#1[[1]] + {1, 0}#2[[1]] #1[[2]] /. {Right -> 1, Left -> 0}),
            #1[[3]],
            dat + (Total@Flatten@{#2[[2]]} /. {Top -> {0, 1}, Bottom -> {0, -1}, Right -> {.3, 0}, Left -> {-.3, 0}})
        ]&,
        {
            Normal@levels[All, {levelConfigPos[#], levelLength[#], Text[Style[typesetConfig[#Configuration, #Term, #J], FontFamily -> "STIXGeneral", 9]]}&],
            pos
        }
    ]
]
allowedDecay[level_] = (#Energy < level[["Energy"]] && Abs[level[["J"]] - #J] <= 1 && !(oddParityQ[Normal@level] && oddParityQ[#] || evenParityQ@Normal@level && evenParityQ@#)&);
wavelength[level_] = (10^7 / Abs[level[["Energy"]] - #Energy]&);
configurationTicks[levelset_] := Normal[{#ConfigurationGroupOrdering, Row[{"          ", Style[typesetConfig@#ConfigurationGroup, FontFamily -> "STIXGeneral", SingleLetterItalics -> False]}, Alignment -> {Left, Bottom}, ImageSize -> {Automatic, 14}]}& /@ levelset[SortBy@"ConfigurationGroupOrdering"][DeleteDuplicatesBy@"ConfigurationGroupOrdering"]]
termSortFunc[term_] := 10ToExpression@#1 + spectroscopicOrder@#2& @@ StringSplit[term, s : LetterCharacter -> s]
NISTLevelPlotLegend[levelset_] := Module[{terms = SortBy[termSortFunc]@Union@StringTrim[DeleteMissing@Normal@levelset[[All, "Term"]], "*"]},
    Scale[MapIndexed[Translate[#1, {0, .45#2[[1]]}]&, {{termColor@#, levelLine[<|"Term" -> #, "J" -> 1 / 2|>]}, Text[Style[typesetTerm@#, FontFamily -> "STIXGeneral", 10, ScriptMinSize -> 7], {.12, 0}]}& /@ Reverse@terms
    ], {2, 1.5}]
]



Begin["`Private`"];

End[];

EndPackage[];